(***********************************************************************)
(*                                                                     *)
(*                                OCaml                                *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* Inclusion checks for the module language *)

ouvre Typedtree
ouvre Types
ouvre Format

val modtypes: Env.t -> module_type -> module_type -> module_coercion
val signatures: Env.t -> signature -> signature -> module_coercion
val compunit: string -> signature -> string -> signature -> module_coercion
val type_declarations:
      Env.t -> Ident.t -> type_declaration -> type_declaration -> unit

type symptom =
    Missing_field de Ident.t * Location.t * string (* kind *)
  | Value_descriptions de Ident.t * value_description * value_description
  | Type_declarations de Ident.t * type_declaration
        * type_declaration * Includecore.type_mismatch list
  | Exception_declarations de
      Ident.t * exception_declaration * exception_declaration
  | Module_types de module_type * module_type
  | Modtype_infos de Ident.t * modtype_declaration * modtype_declaration
  | Modtype_permutation
  | Interface_mismatch de string * string
  | Class_type_declarations de
      Ident.t * class_type_declaration * class_type_declaration *
      Ctype.class_match_failure list
  | Class_declarations de
      Ident.t * class_declaration * class_declaration *
      Ctype.class_match_failure list
  | Unbound_modtype_path de Path.t
  | Unbound_module_path de Path.t

type pos =
    Module de Ident.t | Modtype de Ident.t | Arg de Ident.t | Body de Ident.t
type error = pos list * Env.t * symptom

exception Error de error list

val report_error: formatter -> error list -> unit
val expand_module_alias: Env.t -> pos list -> Path.t -> Types.module_type
